from qkit.core.lib.file_service.breadcrumbs import *
from pytest import fixture
from unittest.mock import patch
import qkit
from pathlib import Path

def test_breadcrumb_name_constant():
    assert derive_breadcrumb_filename() == derive_breadcrumb_filename()

def test_read_single_file():
    datadir_path = Path(__file__).parent
    result = read_breadcrumb(datadir_path / ".1337deadbeef.breadcrumb")
    assert len(result) == 1
    assert result['MNOPQR'] == datadir_path / "file3.h5"

def test_read_all_breadcrumbs():
    datadir_path =  Path(__file__).parent
    result = read_breadcrumbs(datadir_path)
    assert len(result) == 3
    assert result['ABCDEF'] == datadir_path / "file1.h5"
    assert result['GHIJKL'] == datadir_path / "file2.h5"
    assert result['MNOPQR'] == datadir_path / "file3.h5"

@patch("os.getcwd", return_value=(str(Path(__file__).parent)))
@patch("builtins.input", return_value="y")
def test_manual_index(*args, **kwargs):
    # Index the h5 file next to this one.
    qkit.cfg['datadir'] = Path(__file__).parent
    manual_index()
    # Wait on the background task to finish.
    import threading
    [thread.join() for thread in threading.enumerate() if thread.name == "creating_db"]

    # Check indexing worked
    result = read_breadcrumbs(Path(__file__).parent)
    assert result['STUVWX'] == Path(__file__).parent / "STUVWX_dummy.h5"

    # Cleanup breadcrumb
    if derive_breadcrumb_filename().exists():
        import os
        os.remove(derive_breadcrumb_filename())




from pytest import fixture
from pathlib import Path
from collections.abc import Iterable
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from qkit.core.lib.file_service import file_info_database

@fixture(scope="function")
def fid() -> 'Iterable[file_info_database.fid]':
    """
    Load and start qkit, initalize file info database and clean up after test execution
    """
    import qkit
    import qkit.core.s_init.S16_available_modules # Fix class loading
    from qkit.core.lib.file_service import file_info_database
    # Do not initialize with environment variable, patch config
    import os
    if "QKIT_LOCAL_CONFIG" in os.environ:
        del os.environ["QKIT_LOCAL_CONFIG"]
    print(f"Starting with reset cfg.")
    qkit.cfg['datadir'] = Path(__file__).parent
    qkit.cfg['fid_scan_datadir'] = True
    # No qkit.start()
    # Manually create fid:
    fid = file_info_database.fid()
    fid.recreate_database()
    # Wait for indexing to stop
    import threading
    [thread.join() for thread in threading.enumerate() if thread.name == "creating_db"]
    print(f"Indexed: {fid._get_datadir()}")
    print(f"Breadcrumb: {fid._breadcrumb_creator._breadcrumb_file}")
    # Return reference to File Info Data Base. Type error ignorde because of weird qkit import semantics
    yield fid

    # Cleanup afterwards
    from qkit.core.lib.file_service.breadcrumbs import derive_breadcrumb_filename
    print(f"Cleaning up Breadcrumbs at {derive_breadcrumb_filename()}")
    if derive_breadcrumb_filename().exists():
        import os
        os.remove(derive_breadcrumb_filename())


def test_datadir_index(fid: 'file_info_database.fid'):
    # Check Indexing of main directory
    assert Path(fid['STUVWX']) == Path(__file__).parent / "STUVWX_dummy.h5"
    # Check Indexing of sub directories
    assert Path(fid['YZ0123']) == Path(__file__).parent / 'subdir' / "YZ0123_dummy.h5"

def test_creation_time_sorting(fid: 'file_info_database.fid'):
    # Check get_last returning the latest UUID. Note: This is a UUID, not a path!
    assert fid.get_last() == "YZ0123"
    assert Path(fid[fid.get_last()]) == Path(__file__).parent / 'subdir' / "YZ0123_dummy.h5"

def test_adding_file_to_index(fid: 'file_info_database.fid'):
    import os
    datadir = Path(__file__).parent
    path = datadir / "Z45678_dummy.h5"
    with open(path, "w"):
        pass
    # The following is replaced with an internal call to speed up testing:
    # fid.add_h5_file(str(path))
    # import threading
    # from threading import Timer
    # [thread.join() for thread in threading.enumerate() if isinstance(thread, Timer)]
    # The propper call, add_h5_file, is conditional on scanning the data dir and, for some reason,
    # introduces a 20s delay to calling _add. We bypass this by calling directly.
    fid._add(str(path))
    try:
        assert fid.get_last() == "Z45678"
        assert Path(fid[fid.get_last()]) == Path(__file__).parent / "Z45678_dummy.h5"

        # Test Breadcrumb creation on file add
        from qkit.core.lib.file_service.breadcrumbs import read_breadcrumbs
        assert read_breadcrumbs(Path(datadir))['Z45678'] == path
    finally:
        os.remove(path)

def test_get_and__get__equal(fid: 'file_info_database.fid'):
    assert fid.get("YZ0123") == fid["YZ0123"]


from pathlib import Path
from pytest import fixture
import numpy as np
from qkit.storage.store import Data
import tempfile

@fixture(scope="session", autouse=True)
def setup():
    import os
    import pathlib
    from qkit.install.install import create_base_structure
    cwd = pathlib.Path(os.getcwd())
    print("CWD", cwd)
    create_base_structure(cwd)

@fixture
def testdata():
    return np.random.randint(-10, 10, size=(3, 3, 3))

def test_file_creation(testdata):
    with tempfile.TemporaryDirectory() as dir:
        fname = Path(dir) / "hdf5test.h5"

        # Create a test file
        datafile = Data(fname, mode="w")

        x_coord = datafile.add_coordinate("x")
        x_coord.add([0, 1, 2])
        y_coord = datafile.add_coordinate("y")
        y_coord.add([3, 4, 5])
        z_coord = datafile.add_coordinate("z")
        z_coord.add([6, 7, 8])

        # Create a 1D test structure
        dataset1D = datafile.add_value_vector("test1d", x_coord, folder="data")
        dataset1D.append(testdata[0, 0])

        ## Create a 2D test structure
        dataset2D = datafile.add_value_matrix("test2d", x_coord, y_coord, folder="data")
        for x in range(testdata.shape[1]):
            dataset2D.append(testdata[0, x])
        
        # Create a 3D test structure
        dataset3D = datafile.add_value_box("test3d", x_coord, y_coord, z_coord, folder="data")
        for x in range(testdata.shape[0]):
            for y in range(testdata.shape[1]):
                dataset3D.append(testdata[x, y])
            dataset3D.next_matrix() # Breaks without this call

        # Close the file for read back
        datafile.flush()
        datafile.close()

        # Read it back
        datafile = Data(fname)
        assert np.array_equal(datafile.data.test1d[:], testdata[0, 0]), "Failed saving 1D data"
        assert np.array_equal(datafile.data.test2d[:], testdata[0]), "Failed saving 2D data"
        assert np.array_equal(datafile.data.test3d[:], testdata), "Failed saving 3D data"



import os.path
import logging

class QkitCfgError(Exception):
    '''
    If something with qkit.cfg does not fit to what the user wants to do, this is the error to throw.
    '''
    pass

class ConfClass(dict):

    def __init__(self, *args):
        """
        Load configuration from the following sources:
        - Default configuraiton from qkit.config.environment
        - Package overrides from qkit.config.local
        - Configuration file qkit_local_config.py or QKIT_LOCAL_CONFIG environment variable.
        """
        dict.__init__(self, args)

        self._load_default_config()
        sources = self._load_config_overrides() + self._find_and_load_user_config()

        if len(sources) > 1:
            logging.warning("Loaded more than one configuration file! This may cause conflicts! Sources:")
            logging.warning(sources)
        
        # init message
        print ("QKIT configuration initialized -> available as qkit.cfg[...]")
        

    def preset_analyse(self, verbose = False):
        """ Sets basic settings, most of the services are not loaded (default)
            The file index service is run and the UUID registry is populated.
        """

        self['load_info_service'] = False
        self['load_ri_service']   = False
        self['load_visa']         = False
        if verbose:
            print ("Not starting the info_service, ri_service and visa.")

    def preset_measure(self, verbose = False):
        """ Setup of the measurement settings, services are loaded or initialized.
        """
        self['load_info_service'] = True
        self['load_ri_service']   = True
        self['load_visa']         = True
        
        if self['datadir'] == os.path.join(self['qkitdir'],'data'):
                print("Please set a valid data directory! (datadir)")
        if verbose:
            print ("Starting the info_service, ri_service and visa.")
            
    def get(self, item,default=None):
        try:
            return self[item]
        except KeyError:
            if default is not None:
                self[item]=default
            return default

    def _load_default_config(self):
        # load configuration from $QKITDIR/config/*
        try:
            from qkit.config.environment import cfg as cfg_local
            self.update(cfg_local)
        except ImportError:
            pass


    def _load_config_overrides(self) -> list[str]:
        # if a local.py file is defined, load cfg dict and overwrite environment entries.
        _config_sources = []
        try:
            from qkit.config.local import cfg_local
            self.update(cfg_local)
            _config_sources += ["[qkit]src/qkit/config/local.py - cfg_local"]
            logging.warning("DEPRECATED: Loaded qkit.config.local.py! This has been deprecated in favour of qkit_local_config.py, see README!")
        except ImportError:
            pass

        try:    
            from qkit.config.local import cfg as cfg_local
            self.update(cfg_local)
            _config_sources += ["[qkit]src/qkit/config/local.py - cfg"]
            logging.warning("DEPRECATED: Loaded qkit.config.local.py! This has been deprecated in favour of qkit_local_config.py, see README!")
        except ImportError:
            pass
            
        return _config_sources


    def _find_and_load_user_config(self) -> list[str]:
        """
        Loads user configuration from the file system, determined by known file names or environment variables.

        We usually run in a jupyter notebook. If qkit is installed as a package, a local config is not possible,
        and in the long run we should migrate towards a config not in the qkit directory.
        We will use two strategies here:
        1. Check if the environment variable 'QKIT_LOCAL_CONFIG' is set. Use the file it points to.
        2. Use current working directory:
        a. check if a qkit_conf.py file exists. If it does, use it.
        b. Else go to parrent directory and repeat
        If a file was found by either way, load it.

        We log information to stdout, as logging is not yet available. This code is executed before S10_logging.py,
        and will already have become unavailable once logging is initialized.
        """
        global _config_sources
        import os
        from pathlib import Path
        from itertools import chain

        user_config = None
        environment_variable_name = 'QKIT_LOCAL_CONFIG'
        config_file_names = ['qkit_local_config.py', 'local.py']

        if environment_variable_name in os.environ:
            # Inspect environment variable, should it exist
            print("Found environment variable '%s' for configuration." % environment_variable_name)
            user_config = os.environ[environment_variable_name]
        else:
            current_directory = Path(os.getcwd())
            for dir in chain([current_directory], current_directory.parents):
                for (index, config_file_name) in enumerate(config_file_names):
                    test_path = dir / config_file_name
                    if test_path.exists():
                        if test_path.is_file():
                            user_config = test_path
                            if index != 0:
                                print("WARNING: Found legacy configuration file name. Consider migrating.")
                            # We are done here
                            break
                        else:
                            print("WARNING: Found correct file system name %s, but was not a file?!" % test_path)

        # Load if found
        if user_config is not None:
            # https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
            try:
                import importlib.util
                spec = importlib.util.spec_from_file_location("userconfig", user_config)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                self.update(module.cfg)
                return [str(user_config)]
            except Exception as e:
                logging.exception("Could not load user config", e)
        return []



import pytest

import qkit.core

@pytest.fixture(scope="session", autouse=True)
def setup(request):
    import os
    import pathlib
    from qkit.install.install import create_base_structure
    cwd = pathlib.Path(os.getcwd())
    print("CWD", cwd)
    create_base_structure(cwd)

def test_qkit_load():
    import qkit
    qkit.cfg['fid_scan_datadir'] = False
    qkit.cfg['run_id'] = "Basic Test"
    qkit.cfg['user'] = "Automated Test"
    qkit.start()

def test_load_datadir():
    import qkit
    qkit.cfg['fid_scan_datadir'] = True
    qkit.cfg['run_id'] = "Basic Test"
    qkit.cfg['user'] = "Automated Test"
    qkit.start()


# root init for QKIT
# these directories are treated as modules
# HR@KIT/2017/2018
__all__ = ['config','gui','measure','tools', 'analysis','core','instruments','services','storage','logs']


"""
Type hint declarations for easier programming.
TYPE_CHECKING is False at runtime, so it isn't breaking qkit, but True when evaluating variable types
for IntelliSense, making programming easier.
"""
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from qkit.core.lib.file_service.file_info_database import fid as fid_class
    from qkit.config.config_holder import ConfClass
fid: 'fid_class'
cfg: 'ConfClass'

def __getattr__(name):
    """
    Lazy Loading support for qkit configuration. Based on PEP 562: Module __getattr__ and __dir__
    """
    global cfg
    if name == "cfg":
        print("Lazy-Loading configuration...")
        from qkit.config.config_holder import ConfClass
        cfg = ConfClass()
        return cfg
    raise AttributeError("qkit has no attribute " + name)

"""
startup functions
"""
# start initialization (qkit/core/init)
def start(silent=False):
    if not silent:
        print("Starting QKIT framework ... -> qkit.core.startup")
    import qkit.core.startup
    qkit.core.startup.start(silent=silent)


"""
add a few convenience shortcuts 
"""
# remote interface client after qkit.start_ric() -> qkit.ric
from qkit.core.lib.com.ri_client import start_ric
